@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2012 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -59,6 +59,9 @@
 /* for FormatGuard */
 /* #define __NO_FORMATGUARD_ */
 
+/* additional diagnostic messages */
+/* #define DEBUG_FD_ALLOC */
+
 /**************************************** platform */
 
 #ifdef _WIN32
@@ -72,7 +75,6 @@ typedef int socklen_t;
 
 #ifdef USE_WIN32
 #define USE_IPv6
-/* #define USE_FIPS */
 #define _CRT_SECURE_NO_DEPRECATE
 #define _CRT_NONSTDC_NO_DEPRECATE
 #define HAVE_OSSL_ENGINE_H
@@ -162,6 +164,7 @@ typedef int socklen_t;
 #endif
 
 /* must be included before sys/stat.h for Ultrix */
+/* must be included before sys/socket.h for OpenBSD */
 #include <sys/types.h>   /* u_short, u_long */
 /* general headers */
 #include <stdio.h>
@@ -176,6 +179,7 @@ typedef int socklen_t;
 #include <time.h>
 #include <sys/stat.h>    /* stat */
 #include <setjmp.h>
+#include <fcntl.h>
 
 /**************************************** WIN32 headers */
 
@@ -185,6 +189,7 @@ typedef unsigned char u8;
 typedef unsigned short u16;
 typedef unsigned long u32;
 
+#define HAVE_STRUCT_ADDRINFO
 #define HAVE_SNPRINTF
 #define snprintf                    _snprintf
 #define HAVE_VSNPRINTF
@@ -276,19 +281,22 @@ typedef unsigned long u32;
 #endif  /* __vms */
 
     /* Unix-specific headers */
-#include <signal.h>      /* signal */
-#include <sys/wait.h>    /* wait */
+#include <signal.h>         /* signal */
+#include <sys/wait.h>       /* wait */
 #ifdef HAVE_SYS_RESOURCE_H
-#include <sys/resource.h> /* getrlimit */
+#include <sys/resource.h>   /* getrlimit */
 #endif
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>      /* getpid, fork, execvp, exit */
+#include <unistd.h>         /* getpid, fork, execvp, exit */
 #endif
 #ifdef HAVE_STROPTS_H
 #include <stropts.h>
 #endif
+#ifdef HAVE_MALLOC_H
+#include <malloc.h>         /* mallopt */
+#endif
 #ifdef HAVE_SYS_SELECT_H
-#include <sys/select.h>  /* for aix */
+#include <sys/select.h>     /* for aix */
 #endif
 
 #if defined(HAVE_POLL) && !defined(BROKEN_POLL)
@@ -313,7 +321,10 @@ typedef unsigned long u32;
 #ifdef __BEOS__
 #include <posix/grp.h>
 #endif
-#include <fcntl.h>
+
+#ifdef HAVE_SYS_UIO_H
+#include <sys/uio.h>    /* struct iovec */
+#endif /* HAVE_SYS_UIO_H */
 
 #include <netinet/in.h>  /* struct sockaddr_in */
 #include <sys/socket.h>  /* getpeername */
@@ -358,8 +369,12 @@ extern char *sys_errlist[];
 
 /* Linux */
 #ifdef __linux__
+#ifndef IP_FREEBIND
+/* kernel headers without IP_FREEBIND definition */
+#define IP_FREEBIND 15
+#endif /* IP_FREEBIND */
 #ifndef IP_TRANSPARENT
-/* old kernel headers without IP_TRANSPARENT definition */
+/* kernel headers without IP_TRANSPARENT definition */
 #define IP_TRANSPARENT 19
 #endif /* IP_TRANSPARENT */
 #ifdef HAVE_LINUX_NETFILTER_IPV4_H
@@ -405,6 +420,13 @@ extern char *sys_errlist[];
 #endif
 #endif /* HAVE_OSSL_ENGINE_H */
 
+/* non-blocking OCSP API is not available before OpenSSL 0.9.8h */
+#if OPENSSL_VERSION_NUMBER<0x00908080L
+#ifdef HAVE_OSSL_OCSP_H
+#undef HAVE_OSSL_OCSP_H
+#endif /* HAVE_OSSL_OCSP_H */
+#endif /* OpenSSL older than 0.9.8h */
+
 #ifdef HAVE_OSSL_OCSP_H
 #include <openssl/ocsp.h>
 #endif /* HAVE_OSSL_OCSP_H */
@@ -422,6 +444,11 @@ extern char *sys_errlist[];
 #define OPENSSL_NO_TLSEXT
 #endif /* OpenSSL version < 1.0.0 */
 
+#ifndef OPENSSL_NO_COMP
+/* not defined in public headers before OpenSSL 0.9.8 */
+STACK_OF(SSL_COMP) *SSL_COMP_get_compression_methods(void);
+#endif /* OPENSSL_NO_COMP */
+
 /**************************************** other defines */
 
 /* change all non-printable characters to '.' */
@@ -452,6 +479,10 @@ extern char *sys_errlist[];
 #define LOG_DEBUG       7
 #endif /* defined (USE_WIN32) || defined (__vms) */
 
+#ifndef offsetof
+#define offsetof(T, F) ((unsigned int)((char *)&((T *)0L)->F - (char *)0L))
+#endif
+
 #endif /* defined COMMON_H */
 
 /* end of common.h */
