@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2012 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -192,12 +192,12 @@ void libwrap_auth(CLI *c, char *accepted_address) {
         leave_critical_section(CRIT_LIBWRAP);
     }
     if(!result) {
-        s_log(LOG_WARNING, "Service %s REFUSED by libwrap from %s",
+        s_log(LOG_WARNING, "Service [%s] REFUSED by libwrap from %s",
             c->opt->servname, accepted_address);
         s_log(LOG_DEBUG, "See hosts_access(5) manual for details");
         longjmp(c->err, 1);
     }
-    s_log(LOG_DEBUG, "Service %s permitted by libwrap from %s",
+    s_log(LOG_DEBUG, "Service [%s] permitted by libwrap from %s",
         c->opt->servname, accepted_address);
 }
 
