@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2012 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -101,13 +101,15 @@ int s_poll_canwrite(s_poll_set *fds, int fd) {
     return 0;
 }
 
-int s_poll_error(s_poll_set *fds, int fd) {
+int s_poll_error(s_poll_set *fds, FD *s) {
     unsigned int i;
 
+    if(!s->is_socket)
+        return 0;
     for(i=0; i<fds->nfds; i++)
-        if(fds->ufds[i].fd==fd)
+        if(fds->ufds[i].fd==s->fd)
             return fds->ufds[i].revents&(POLLERR|POLLNVAL) ?
-                get_socket_error(fd) : 0;
+                get_socket_error(s->fd) : 0;
     return 0;
 }
 
@@ -296,6 +298,7 @@ void s_poll_free(s_poll_set *fds) {
 void s_poll_init(s_poll_set *fds) {
     FD_ZERO(&fds->irfds);
     FD_ZERO(&fds->iwfds);
+    FD_ZERO(&fds->ixfds);
     fds->max=0; /* no file descriptors */
 }
 
@@ -304,6 +307,8 @@ void s_poll_add(s_poll_set *fds, int fd, int rd, int wr) {
         FD_SET((unsigned int)fd, &fds->irfds);
     if(wr)
         FD_SET((unsigned int)fd, &fds->iwfds);
+    /* always expect errors (and the Spanish Inquisition) */
+    FD_SET((unsigned int)fd, &fds->ixfds);
     if(fd>fds->max)
         fds->max=fd;
 }
@@ -316,10 +321,14 @@ int s_poll_canwrite(s_poll_set *fds, int fd) {
     return FD_ISSET(fd, &fds->owfds);
 }
 
-int s_poll_error(s_poll_set *fds, int fd) {
-    if(!FD_ISSET(fd, &fds->orfds)) /* error conditions are signaled as read */
+int s_poll_error(s_poll_set *fds, FD *s) {
+    if(!s->is_socket)
+        return 0; /* getsockopt is only available on sockets */
+    /* error conditions are signaled as read, but apparently *not* in Winsock:
+     * http://msdn.microsoft.com/en-us/library/windows/desktop/ms737625%28v=vs.85%29.aspx */
+    if(!(FD_ISSET(s->fd, &fds->orfds) || FD_ISSET(s->fd, &fds->oxfds)))
         return 0;
-    return get_socket_error(fd); /* check if it's really an error */
+    return get_socket_error(s->fd); /* check if it's really an error */
 }
 
 int s_poll_wait(s_poll_set *fds, int sec, int msec) {
@@ -329,6 +338,7 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
     do { /* skip "Interrupted system call" errors */
         memcpy(&fds->orfds, &fds->irfds, sizeof(fd_set));
         memcpy(&fds->owfds, &fds->iwfds, sizeof(fd_set));
+        memcpy(&fds->oxfds, &fds->ixfds, sizeof(fd_set));
         if(sec<0) { /* infinite timeout */
             tv_ptr=NULL;
         } else {
@@ -336,7 +346,7 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
             tv.tv_usec=1000*msec;
             tv_ptr=&tv;
         }
-        retval=select(fds->max+1, &fds->orfds, &fds->owfds, NULL, tv_ptr);
+        retval=select(fds->max+1, &fds->orfds, &fds->owfds, &fds->oxfds, tv_ptr);
     } while(retval<0 && get_last_socket_error()==S_EINTR);
     return retval;
 }
@@ -350,8 +360,9 @@ int set_socket_options(int s, int type) {
     extern SOCK_OPT sock_opts[];
     static char *type_str[3]={"accept", "local", "remote"};
     int opt_size;
+    int retval=0; /* no error found */
 
-    for(ptr=sock_opts;ptr->opt_str;ptr++) {
+    for(ptr=sock_opts; ptr->opt_str; ptr++) {
         if(!ptr->opt_val[type])
             continue; /* default */
         switch(ptr->opt_type) {
@@ -376,14 +387,17 @@ int set_socket_options(int s, int type) {
                     ptr->opt_str, type_str[type]);
             } else {
                 sockerror(ptr->opt_str);
-                return -1; /* FAILED */
+                retval=-1; /* failed to set this option */
             }
-        } else {
+        }
+#ifdef DEBUG_FD_ALLOC
+        else {
             s_log(LOG_DEBUG, "Option %s set on %s socket",
                 ptr->opt_str, type_str[type]);
         }
+#endif /* DEBUG_FD_ALLOC */
     }
-    return 0; /* OK */
+    return retval; /* returns 0 when all options succeeded */
 }
 
 int get_socket_error(const int fd) {
@@ -391,7 +405,7 @@ int get_socket_error(const int fd) {
     socklen_t optlen=sizeof err;
 
     if(getsockopt(fd, SOL_SOCKET, SO_ERROR, (void *)&err, &optlen))
-        return get_last_socket_error(); /* failed -> ask why */
+        err=get_last_socket_error(); /* failed -> ask why */
     return err;
 }
 
@@ -434,16 +448,12 @@ int connect_blocking(CLI *c, SOCKADDR_UNION *addr, socklen_t addrlen) {
         str_free(dst);
         return -1;
     default:
-        if(s_poll_canread(c->fds, c->fd) || s_poll_error(c->fds, c->fd)) {
-            /* newly connected socket should not be ready for read */
-            /* get the resulting error code, now */
-            error=get_socket_error(c->fd);
-            if(error) { /* really an error? */
-                s_log(LOG_ERR, "connect_blocking: getsockopt %s: %s (%d)",
-                    dst, s_strerror(error), error);
-                str_free(dst);
-                return -1;
-            }
+        error=get_socket_error(c->fd);
+        if(error) {
+            s_log(LOG_ERR, "connect_blocking: connect %s: %s (%d)",
+               dst, s_strerror(error), error);
+            str_free(dst);
+            return -1;
         }
         if(s_poll_canwrite(c->fds, c->fd)) {
             s_log(LOG_NOTICE, "connect_blocking: connected %s", dst);
